
import xarray as xr
from fsspec import AbstractFileSystem
from fsspec.asyn import AsyncFileSystem

from typing import Union
from zarr.storage import FSStore


def open_cube(path: str, fs: Union[AbstractFileSystem, AsyncFileSystem] = None, group: str = None, mode: str = "r", **kwargs):

    store = FSStore(path, mode=mode, fs=fs)

    return xr.open_zarr(store, group=group, **kwargs)




from typing import Optional, Union

import numpy as np
import pandas as pd
import shapely
from shapely.geometry.polygon import Polygon
from shapely.geometry.multipolygon import MultiPolygon


class Bbox:
    """"Class for bounding box represented by four number.

    Args:
        minx (Union[str, int, float]): Minimum coordinate in x direction.
        miny (Union[str, int, float]): Minimum coordinate in y direction.
        maxx (Union[str, int, float]): Maximum coordinate in x direction.
        maxy (Union[str, int, float]): Maximum coordinate in y direction.
        precision (int): presion applied at rounding.
    """

    def __init__(self,
                 minx: Union[str, int, float],
                 miny: Union[str, int, float],
                 maxx: Union[str, int, float],
                 maxy: Union[str, int, float],
                 precision: Optional[int] = None):

        box = [minx, miny, maxx, maxy]
        if any([isinstance(v, str) for v in box]):
            box = list(map(float, box))

        if precision:
            box[0] = np.round(box[0], precision)
            box[1] = np.round(box[1], precision)
            box[2] = np.round(box[2], precision)
            box[3] = np.round(box[3], precision)

        self._minx = box[0]
        self._miny = box[1]
        self._maxx = box[2]
        self._maxy = box[3]
        self._precision = precision

    def __dict__(self):
        return {'minx': self.minx,
                'miny': self.miny,
                'maxx': self.maxx,
                'maxy': self.maxy}

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):

        if len(v) == 4:
            precision = None
        elif len(v) == 5:
            precision = v[-1]
        else:
            raise ValueError(
                f'Invalid number of parameters of {len(v)} - bbox object '
                'must consist of four coordinate numbers and optionally '
                'a precision number.')

        try:
            minx, miny, maxx, maxy = tuple(map(float, (v[0], v[1],
                                                       v[2], v[3])))
            if not precision:
                base_prec = sorted(
                    map(str, (minx, miny, maxx, maxy)), key=len)[-1]
                precision = len(base_prec.split('.')[1])
            else:
                precision = int(precision)
        except (TypeError, ValueError) as error:
            raise ValueError(
                'Invalid input parameters - coordinates and precision '
                'must be convertable to numbers.') from error

        if minx > maxx:
            raise ValueError('minx must be smaller or equal to maxx')

        if miny > maxy:
            raise ValueError('miny must be smaller or equal to maxy')

        minx, miny, maxx, maxy = tuple(
            map(np.round, (minx, miny, maxx, maxy), [precision]*4))

        return cls(minx, miny, maxx, maxy, precision)

    @property
    def bbox(self):
        """Return the bounding box as a tuple."""
        return (self._minx, self._miny, self._maxx, self._maxy)

    @property
    def minx(self):
        """Return minimum value in x direction."""
        return self._minx

    @minx.setter
    def minx(self, value):
        """Set minimum value in x direction."""
        if self._precision:
            self._minx = np.round(float(value), self._precision)
        else:
            self._minx = float(value)

    @property
    def miny(self):
        """Return minimum value in y direction."""
        return self._miny

    @miny.setter
    def miny(self, value):
        """Set minimum value in y direction."""
        if self._precision:
            self._miny = np.round(float(value), self._precision)
        else:
            self._miny = float(value)

    @property
    def maxx(self):
        """Return maximum value in x direction."""
        return self._maxx

    @maxx.setter
    def maxx(self, value):
        """Set maximum value in x direction."""
        if self._precision:
            self._maxx = np.round(float(value), self._precision)
        else:
            self._maxx = float(value)

    @property
    def maxy(self):
        """Return maximum value in y direction."""
        return self._maxy

    @maxy.setter
    def maxy(self, value):
        """Set maximum value in y direction."""
        if self._precision:
            self._maxy = np.round(float(value), self._precision)
        else:
            self._maxy = float(value)


class TimeRange:
    """"Class for time range represented by two dates.

    Args:
        start_time (Union[str, pd.Timestamp]): Start date.
        end_time (Union[str, pd.Timestamp]): End date.
    """

    def __init__(self,
                 start_time: Union[str, pd.Timestamp],
                 end_time: Union[str, pd.Timestamp]):

        start_time = self.convert_time(start_time)
        end_time = self.convert_time(end_time)

        if start_time > end_time:
            raise ValueError("start_time must be smaller or equal to end_time")

        self._start_time = start_time
        self._end_time = end_time

    def __repr__(self):
        obj_name = f'<ddc_utility.{type(self).__name__}>'
        rep = f"{obj_name}\nstart_time: {self.start_time}\nend_time: {self.end_time}\n"
        return rep

    @property
    def time_range(self):
        """Return time range."""
        return [self._start_time, self._end_time]

    @property
    def start_time(self):
        """Return start time."""
        return self._start_time

    @start_time.setter
    def start_time(self, value):
        """Set start time."""
        start_time = self.convert_time(value)
        if start_time <= self._end_time:
            self._start_time = start_time
        else:
            raise ValueError('start_time must be smaller or '
                             'equal to end_time')

    @property
    def end_time(self):
        """Return end time."""
        return self._end_time

    @end_time.setter
    def end_time(self, value):
        """Set start time."""
        end_time = self.convert_time(value)
        if end_time >= self._start_time:
            self._end_time = end_time
        else:
            raise ValueError('end_time must be greater or '
                             'equal to start_time')

    def to_string(self, only_date=True):
        """Return custom string representation of the time range.

        Args:
            only_date (bool): Wheter to return only the date part of time range.
                Defaults to True.
        """
        if only_date:
            return [self._start_time.isoformat(sep='T').split('T')[0],
                    self._end_time.isoformat(sep='T').split('T')[0]]

        return [self._start_time.isoformat(sep='T'),
                self._end_time.isoformat(sep='T')]

    def convert_to_full_months(self):
        """
        Convert date to full months.
        For example, in the case of start time '2021-01-04' to '2021-01-01' or
        in the case of end time '2021-05-15' to '2021-05-31'.
        """
        self._start_time = self._start_time.replace(day=1)
        self._end_time = (self._end_time if self._end_time.is_month_end
                          else self._end_time + pd.offsets.MonthEnd())

    @classmethod
    def convert_time(cls,
                     datetime: Union[str, pd.Timestamp],
                     utc: bool = False):
        """Convert time to pandas Timestamp."""
        try:
            return pd.to_datetime(datetime, utc=utc)
        except Exception as error:
            raise ValueError(
                "Invalid input parameters -- must be convertable to pandas.TimeStamp") from error


class Geometry:
    """"Class for WKT type geometry.

        Args:
            geometry (Union[str, Polygon, MultiPolygon]): Geometry.
    """

    def __init__(self,
                 geometry: Union[str, Polygon, MultiPolygon]):

        if not isinstance(geometry, (Polygon, MultiPolygon)):
            geometry = self.convert_geometry(geometry)

        if not geometry.is_valid:
            raise ValueError(
                "Invalid geometry -- must be valid based on shapely's definition")

        self._geometry = geometry

    def __repr__(self):
        obj_name = f'<ddc_utility.{type(self).__name__}>'
        rep = f"{obj_name}\ngeometry: {self._geometry}\n"
        return rep

    @property
    def geometry(self):
        """Return geometry."""
        return self._geometry

    def to_string(self):
        """Return WKT string representation of the geometry."""
        return self._geometry.wkt

    @classmethod
    def convert_geometry(cls,
                         geometry: str):
        """Convert geometry to Shapely object."""
        try:
            geometry = shapely.from_wkt(geometry)
        except shapely.errors.GEOSException:
            geometry = MultiPolygon(shapely.from_geojson(geometry))
        except Exception as error:
            raise ValueError(
                "Invalid input parameters -- must be convertable to Shapely.geometry") from error

        if geometry.geom_type not in ['Polygon', 'MultiPolygon']:
            raise ValueError(
                "Invalid geometry type -- must be Polygon or MultiPolygon")

        return geometry


class IrrigationSchedule:
    """"Class for defining irrigation schedule.

        Args:
            date (Union[str, pd.Timestamp]): Irrigation date.
            amount (float): Amount of irrigated water in mm.
    """

    def __init__(self, date: Union[str, pd.Timestamp], amount: float):

        try:
            date = pd.to_datetime(date)
        except Exception as error:
            raise ValueError(
                "Invalid input parameters -- 'date' must be convertable to pandas.TimeStamp") from error

        try:
            amount = float(amount)
        except ValueError as error:
            raise ValueError(
                "Invalid input parameters -- 'amount must be convertable to float") from error

        self._date = pd.to_datetime(date)
        self._amount = float(amount)

    def __repr__(self):
        obj_name = f'<ddc_utility.{type(self).__name__}>'
        rep = f"{obj_name}\ndate: {self._date}\namount: {self._amount}\n"
        return rep

    @property
    def date(self):
        """Return geometry."""
        return self._date

    @property
    def amount(self):
        """Return geometry."""
        return self._amount

    def to_string(self, only_date=True):
        """Return the tuple string representation of the object."""
        if only_date:
            return (self._date.isoformat(sep='T').split('T')[0], self._amount)
        else:
            return (self._date.isoformat(sep='T'), self._amount)



from pydantic import BaseModel, model_validator, field_validator, computed_field
from typing import Optional, List, Dict


DDC_DATASETS = {
    'HR_VPP_DAILY': {'title': 'Daily Vegetation Phenology and Productivity', 'bands': ['LAI', 'NDVI', 'FAPAR', 'QFLAG2', 'PPI']},
    'HR_VPP_ST': {'title': '10-daily Seasonal Trajectories', 'bands': ['PPI', 'QFLAG']},
    'S2L2A': {'title': 'Sentinel-2', 'bands': ['B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B08', 'B8A', 'B09', 'B11', 'B12', 'CLD']},
    'S1GRD': {'title': 'Sentinel-1 GRD', 'bands': ['VV', 'VH', 'localIncidenceAngle', 'scatteringArea', 'shadowMask']},
    'Meteorology': {'title': 'Meteorology', 'bands': ['ET0', 'Temp Avg', 'Temp Min', 'Temp Max', 'Prec']},
    'Soil': {'title': 'Soil', 'bands': ['hyd_cur_K0', 'hyd_cur_L', 'hyd_cur_alp', 'hyd_cur_m', 'hyd_cur_n', 'hyd_cur_thr', 'hyd_cur_ths', 'hydcon', 'moi_cur_alp', 'moi_cur_m', 'moi_cur_n', 'moi_cur_thr', 'moi_cur_ths', 'watcap', 'watcnt', 'watwil']},
    'Nhrl': {'title': 'Nhrl', 'bands': ['nhrl']},
    'Ecosystem': {'title': 'Ecosystem', 'bands': ['ecosystem']},
    'S2L2A_PC': {'title': 'Sentinel-2 PC', 'bands': ['B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B08', 'B8A', 'B09', 'B11', 'B12']}
}


class DataCollection(BaseModel):

    # title: str
    name: str
    bands: Optional[List[str]]

    @field_validator('name')
    @classmethod
    def check_name(cls, val):
        if val not in DDC_DATASETS:
            raise ValueError(
                f'Invalid dataset name, name must be in {list(DDC_DATASETS)}')
        return val

    @model_validator(mode="after")
    def check_bands(self):
        orig_bands = DDC_DATASETS[self.name]["bands"]
        new_bands = []

        for band in self.bands:
            if band in orig_bands:
                new_bands.append(band)
        self.bands = new_bands
        return self

    @computed_field
    @property
    def title(self) -> str:
        return DDC_DATASETS[self.name]["title"]


def prepare_data_layers(data_layers: List[DataCollection]) -> List[Dict]:

    layers = []
    available_datasets = list(DDC_DATASETS)

    for collection in data_layers:
        layers.append(collection.model_dump())
        available_datasets.remove(collection.name)

    for rest in available_datasets:
        layers.append(DataCollection(name=rest, bands=[]).model_dump())

    return layers




from typing import Dict, List, Optional


class DanubeDataCubeMetadata:

    def __init__(self):
        self._metadata = _DDC_METADATA

    @property
    def dataset_names(self) -> List[str]:
        return [ds_id for ds_id in self._metadata['datasets']]

    def dataset(self, dataset_name: str) -> Optional[Dict]:
        dataset = dict(self._dataset_direct(dataset_name))
        dataset.pop('variables')
        return dataset if dataset is not None else None

    def dataset_vars(self, dataset_name: str, default=None) -> \
            Optional[Dict]:
        variables = self._dataset_vars_direct(dataset_name)
        return dict(variables) if variables else default

    def dataset_vars_names(self, dataset_name: str, default=None) -> \
            Optional[List[str]]:
        variables = self._dataset_vars_direct(dataset_name)
        return [var_name for var_name in variables] if variables else default

    def dataset_var(self, dataset_name: str, variable_name: str, default=None) \
            -> Optional[Dict]:
        var = self._dataset_var_direct(dataset_name, variable_name)
        return dict(var) if var else default

    def _dataset_direct(self, dataset_name: str) -> Optional[Dict]:
        return self._metadata['datasets'].get(dataset_name.upper())

    def _dataset_vars_direct(self, dataset_name: str) \
            -> Optional[Dict]:
        dataset = self._dataset_direct(dataset_name)
        return dataset.get('variables') if dataset else {}

    def _dataset_var_direct(self, dataset_name: str, variable_name: str) \
            -> Optional[Dict]:
        variables = self._dataset_vars_direct(dataset_name)
        return variables.get(variable_name) if variables else {}


_FORECAST_VAR_METADATA = {
    'temp_min': {
        'standard_name': 'min_air_temperature_2m',
        'long_name': 'Minimum temperature (2 m)',
        'units': '°C',
        'coverage_content_type': 'physical_measurement',
    },
    'temp_max': {
        'standard_name': 'max_air_temperature_2m',
        'long_name': 'Maximum temperature (2 m)',
        'units': '°C',
        'coverage_content_type': 'physical_measurement',
    },
    'temp_avg': {
        'standard_name': 'avg_air_temperature_2m',
        'long_name': 'Average temperature (2 m)',
        'units': '°C',
        'coverage_content_type': 'physical_measurement',
    },
    'prec': {
        'standard_name': 'precipitation_mm',
        'long_name': 'Precipitation amount',
        'units': 'mm',
        'coverage_content_type': 'physical_measurement',
    },
    'et0': {
        'standard_name': 'evapotranspiration_mm',
        'long_name': 'Potential evapotranspiration',
        'units': 'mm',
        'coverage_content_type': 'physical_measurement',
    }
}


_MET_VAR_METADATA = {
    'temp_min': {
        'standard_name': 'min_air_temperature_2m',
        'long_name': 'Minimum temperature (2 m)',
        'units': '°C',
        'coverage_content_type': 'physical_measurement',
    },
    'temp_max': {
        'standard_name': 'max_air_temperature_2m',
        'long_name': 'Maximum temperature (2 m)',
        'units': '°C',
        'coverage_content_type': 'physical_measurement',
    },
    'temp_avg': {
        'standard_name': 'avg_air_temperature_2m',
        'long_name': 'Average temperature (2 m)',
        'units': '°C',
        'coverage_content_type': 'physical_measurement',
    },
    'prec': {
        'standard_name': 'precipitation_mm',
        'long_name': 'Precipitation amount',
        'units': 'mm',
        'coverage_content_type': 'physical_measurement',
    },
    'eto': {
        'standard_name': 'evapotranspiration_mm',
        'long_name': 'Potential evapotranspiration',
        'units': 'mm',
        'coverage_content_type': 'physical_measurement',
    }
}


_SOIL_VAR_METADATA = {
    'soil1': {
        'standard_name': '',
        'long_name': '',
        'units': "",
        'coverage_content_type': '',
    }
}


_NHRL_VAR_METADATA = {
    'nhrl': {
        'standard_name': 'landcover_categories',
        'long_name': 'Land cover and land use thematic map',
        'units': "thematic class/category",
        'coverage_content_type': 'thematic_classification',
        'flag_meanings': '["other_paved_or_non-paved_artificial_surfaces", "road_network", "railway_network", "bare_soil", "cereals", "row_crops", "alfalfa", "rapeseed", "grasslands", "orchards", "vineyards", "heterogeneous_vegetated_surfaces_in_built-up_environment", "shrubland", "broadleaved_trees", "coniferous_trees", "grasslands_with_shrubs", "heterogeneous_grassland_with_periodic_water_effect", "salt_steppes_and_meadows", "wetlands", "reeds", "water_surfaces", "single-floor_buildings", "high_buildings"]',
        'flag_values': '[12, 13, 14, 20, 21, 22, 24, 25, 26, 27, 28, 29, 30, 31, 33, 35, 36, 37, 41, 43, 51, 111, 112]',
        'flag_colors': '["#828282", "#4e4e4e", "#000000", "#e1e1e1", "#ffffbe", "#ffebaf", "#d3ffbe", "#aaff00", "#abcd66", "#b0643c", "#ffaa00", "#ffff00", "#448970", "#38a800", "#267300", "#f5f57a", "#cdf57a", "#67f070", "#73b2ff", "#d69dbc", "#0070ff", "#ff0000", "#a80000"]',
    }
}

_ECOSYSTEM_VAR_METADATA = {
    'ecosystem': {
        'standard_name': 'ecosystem_categories',
        'long_name': 'Ecosystem categories thematic map',
        'units': "thematic class/category",
        'coverage_content_type': 'thematic_classification',
        'flag_meanings': '["low_buildings", "high_buildings", "paved_roads", "dirt_roads", "railways", "other_paved_or_non-paved_artificial_areas", "green_urban_areas_with_trees", "green_urban_areas_without_trees", "arable_land", "vineyards", "fruit_and_berry_and_other_plantations", "energy_crops", "complex_cultivation_patterns_with_scattered_buildings", "complex_cultivation_patterns_without_buildings", "open_sand_steppes", "closed_sand_steppes", "salt_steppes_and_meadows_(grasslands_affected_by_salinisation_included)", "calcareous_open_rocky_grasslands", "siliceous_open_rocky_grasslands", "closed_grasslands_in_hills_and_mountains_or_on_cohesive_soil", "other_herbaceous_vegetation", "beech_forests", "sessile_oak-hornbeam_forests", "turkey_oak_forests", "downy_oak_forests", "scots_pine_stands_of_western_transdanubia", "deciduous_stands_of_western_transdanubia_mixed_with_scots_pine", "native_poplar_dominated_forests", "pioneer_forests_of_hilly_and_mountainous_regions", "pedunculate_oak-hornbeam_forests", "pedunculate_oak_forests_monospecific_or_mixed_with_ash", "forests_dominated_by_other_native_tree_species", "other_mixed_deciduous_forests", "riverine_willow-poplar_woodlands", "riverine_hardwood_forests", "pedunculate_oak_forests_monospecific_or_mixed_with_ash", "alder_forests", "pedunculate_oak-hornbeam_forests_(with_excess_water)", "willow_woods_outside_the_floodplain", "poplar_woods_outside_the_floodplain", "birch_woodland", "turkey_oak_forests_with_excess_water", "forests_dominated_by_other_native_tree_species_with_excess_water", "other_mixed_deciduous_forests_with_excess_water", "conifer-dominated_plantations", "black_locust-dominated_mixed_plantations", "plantations_dominated_by_non-native_poplar_and_willow_species", "plantations_of_other_non-native_tree_species", "clearcut", "forest_stand_under_regeneration", "other_ligneous_vegetation_woodlands", "tall-herb_vegetation_of_marshes_and_fens_standing_in_water", "fens_and_mesotrophic_wet_meadows_grasslands_with_periodic_water_effect", "swamp_woodlands", "water_bodies", "water_courses"]',
        'flag_values': '[1110, 1120, 1210, 1220, 1230, 1310, 1410, 1420, 2100, 2210, 2220, 2230, 2310, 2320, 3110, 3120, 3200, 3310, 3320, 3400, 3500, 4101, 4102, 4103, 4104, 4105, 4106, 4107, 4108, 4109, 4110, 4111, 4112, 4201, 4202, 4301, 4302, 4303, 4304, 4305, 4306, 4307, 4308, 4309, 4401, 4402, 4403, 4404, 4501, 4502, 4600, 5110, 5120, 5200, 6100, 6200]',
        'flag_colors': '["#FF0000", "#A80000", "#9C9C9C", "#CCCCCC", "#828282", "#FFBEBE", "#A8A800", "#D7D79E", "#FFEBBE", "#FFD37F", "#FFAA00", "#E69800", "#FFB582", "#FFFF73", "#D1FF73", "#D1FF73", "#CFBBA7", "#C8DC87", "#C8DC87", "#A1B36D", "#98E600", "#38A800", "#38A800", "#38A800", "#38A800", "#38A800", "#38A800", "#38A800", "#38A800", "#38A800", "#38A800", "#38A800", "#70A800", "#70A800", "#50B400", "#50B400", "#50B400", "#50B400", "#50B400", "#50B400", "#50B400", "#50B400", "#50B400", "#507300", "#507300", "#507300", "#507300", "#737300", "#737300", "#267300", "#BEE8FF", "#73DFFF", "#00C5FF", "#004DA8", "#002673"]',
    }
}


_DDC_METADATA = dict(
    datasets={
        'FORECAST': dict(
            title='OMSZ Meteorological Forecast',
            description='Interpolated grids of meteorological forecast.',
            variables=_FORECAST_VAR_METADATA
        ),
        'METEOROLOGY': dict(
            title='OMSZ Meteorological Measurements',
            description='Interpolated grids of meteorological measurements.',
            variables=_MET_VAR_METADATA
        ),
        'SOIL': dict(
            title='Soil data',
            description='',
            variables=_SOIL_VAR_METADATA
        ),
        'NHRL': dict(
            title='National High-Resolution Layer of Hungary (NHRL)',
            description='Derived from EO satellite imagery, GIS data and in-situ measurements. The data generation is based on Random Forest machine learning algorithm.',
            copyright='Lechner Nonprofit Ltd.',
            institution='Lechner Nonprofit Ltd., Directorate of Geospatial Services',
            licence='This work is licensed under a Creative Commons Attribution 4.0 International License (https://creativecommons.org/licenses/by/4.0/).',
            variables=_NHRL_VAR_METADATA,
        ),
        'ECOSYSTEM': dict(
            title='Ecosystem Map of Hungary',
            description='Derived from EO satellite imagery, GIS data and in-situ measurements. The data generation is based on Random Forest machine learning algorithm.',
            copyright='Hungarian Ministry of Agriculture',
            institution='Lechner Nonprofit Ltd., Directorate of Geospatial Services',
            licence='The downloaded databases are free to use as long as references to the source are provided. Citacion (obligatory) - The database/analysis has been created using the Ecosystem Map of Hungary (project KEHOP-430-VEKOP-15-2016-00001, Ministry of Agriculture, 2019) DOI - 10.34811/osz.alapterkep',
            variables=_ECOSYSTEM_VAR_METADATA,

        ),
    }
)



import pytest
from ddc_utility.metadata import DanubeDataCubeMetadata


@pytest.fixture
def meta():
    return DanubeDataCubeMetadata()


def test_dataset_names(meta):
    assert set(meta.dataset_names) == {
        'METEOROLOGICAL_FORECAST', 'METEOROLOGICAL', 'SOIL', 'NHRL', 'ECOSYSTEM'}


@pytest.mark.parametrize("dataset_name, expected", [
    ('FORECAST', dict),
    ('HISTORICAL', dict),
    ('DAILY', dict),
    ('SOIL1', dict),
    ('NHRL', dict),
    ('ECOSYSTEM', dict),
    ('not_valid', type(None)),
])
def test_dataset_type(meta, dataset_name, expected):
    assert type(meta.dataset(dataset_name)) == expected


def test_dataset_typeerror(meta):
    with pytest.raises(TypeError):
        meta.dataset()


@pytest.mark.parametrize("dataset_name, expected", [
    ('FORECAST', 'WGS84 (EPSG:4326)'),
    ('HISTORICAL', 'WGS84 (EPSG:4326)'),
    ('DAILY', 'WGS84 (EPSG:4326)'),
    ('SOIL1', 'WGS84 (EPSG:4326)'),
    ('NHRL', 'WGS84 (EPSG:4326)'),
    ('ECOSYSTEM', 'WGS84 (EPSG:4326)'),
])
def test_dataset_crs(meta, dataset_name, expected):
    assert meta.dataset('NHRL').get('delivered_product_crs') == expected


@pytest.mark.parametrize("dataset_name, expected", [
    ('FORECAST', 'meteorological'),
    ('HISTORICAL', 'meteorological'),
    ('DAILY', 'meteorological'),
    ('SOIL1', 'soil'),
    ('NHRL', 'environmental'),
    ('ECOSYSTEM', 'environmental'),
])
def test_dataset_collection_name(meta, dataset_name, expected):
    assert meta.dataset_collection_name(dataset_name) == expected


def test_dataset_vars(meta):
    assert set(meta.dataset_vars('FORECAST').keys()) == {
        'precipitation', 'temperature_max', 'temperature_min'}
    with pytest.raises(AttributeError):
        meta.dataset_vars('not_valid').keys()


def test_dataset_vars_names(meta):
    assert set(meta.dataset_vars_names('FORECAST')) == {
        'precipitation', 'temperature_max', 'temperature_min'}


def test_dataset_var(meta):
    assert meta.dataset_var('NHRL', 'nhrl').get(
        'standard_name') == 'landcover_categories'
    with pytest.raises(AttributeError):
        meta.dataset_var('NHRL', 'not_valid').get('standard_name')


@pytest.mark.parametrize("collection_name, expected", [
    ('meteorological', {'DAILY', 'FORECAST', 'HISTORICAL'}),
    ('soil', {'SOIL1'}),
    ('environmental', {'ECOSYSTEM', 'NHRL'}),
])
def test_dataset_by_collection_name(meta, collection_name, expected):
    assert set(meta.dataset_by_collection_name(
        collection_name).keys()) == expected



import logging
import os
import platform
import time
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd
import requests
import s3fs
import xarray as xr
from shapely.geometry.multipolygon import MultiPolygon
from shapely.geometry.polygon import Polygon

from ddc_utility.__version__ import version
from ddc_utility.auth import OAuth2BearerHandler
from ddc_utility.constants import (DEFAULT_AOI_BUCKET, DEFAULT_DDC_BUCKET,
                                   DEFAULT_DDC_HOST)
from ddc_utility.cube import open_cube
from ddc_utility.data_collections import DataCollection, prepare_data_layers, DDC_DATASETS
from ddc_utility.errors import (BadRequest, DdcClientError, DdcException,
                                DdcRequestError, Forbidden, HTTPException,
                                NotFound, ServerError, TooManyRequests,
                                Unauthorized)
from ddc_utility.utils import Geometry, IrrigationSchedule, TimeRange

log = logging.getLogger(__name__)


def authorize_request(method):
    def wrapper(self, *args, **kwargs):
        now = round(time.time())
        if self.auth is None or (self.auth.expires_at - now < 60):
            token = self.fetch_token()
            self.auth = OAuth2BearerHandler(
                token.get('access_token'), token.get('expires_at'))

        return method(self, *args, **kwargs)
    return wrapper


def authorize_s3_access(method):
    def wrapper(self, *args, **kwargs):
        now = round(time.time())

        if self.aws_s3 is None or (self.aws_session_exp - now < 60):
            temp_cred = self.fetch_temporary_credentials()
            self.aws_s3 = s3fs.S3FileSystem(
                key=temp_cred["AccessKeyId"],
                secret=temp_cred["SecretKey"],
                token=temp_cred["SessionToken"])
            self.aws_session_exp = temp_cred["expires_at"]

        return method(self, *args, **kwargs)
    return wrapper


class BaseClient:

    def __init__(
            self,
            client_id: Optional[str] = None,
            client_secret: Optional[str] = None,
            host: Optional[str] = None,
            wait_on_rate_limit: Optional[bool] = False) -> None:

        self.client_id = client_id
        self.client_secret = client_secret
        self.host = host

        self.wait_on_rate_limit = wait_on_rate_limit

        self.session = requests.Session()
        self.user_agent = (
            f"Python/{platform.python_version()} "
            f"Requests/{requests.__version__} "
            f"ddc_cube/{version}"
        )

    def request(self, method: str, route: str, params: Optional[Dict] = None,
                json_data: Optional[Dict] = None, auth: requests.auth.AuthBase = None, content_type: Optional[str] = None, accept: Optional[str] = None):

        headers = {
            "User-Agent": self.user_agent,
            "client_id": self.client_id
        }
        if content_type is not None:
            headers["Content-Type"] = content_type
        if accept is not None:
            headers["Accept"] = accept

        url = self.host + route

        log.debug(
            f"\nMaking API request: {method} {url}\n"
            f"Parameters: {params}\n"
            f"Headers: {headers}\n"
            f"Body: {json_data}"
        )

        with self.session.request(
                method, url, params=params, json=json_data, headers=headers, auth=auth) as response:

            if isinstance(response.content, bytes):
                log.debug(
                    "\nReceived API response: "
                    f"{response.status_code} {response.reason}\n"
                    f"Headers: {response.headers}\n"
                    f"Content: 'bytes content'"
                )

            else:
                log.debug(
                    "\nReceived API response: "
                    f"{response.status_code} {response.reason}\n"
                    f"Headers: {response.headers}\n"
                    f"Content: {response.content}"
                )

            if response.status_code == 400:
                raise BadRequest(response)
            if response.status_code == 401:
                raise Unauthorized(response)
            if response.status_code == 403:
                raise Forbidden(response)
            if response.status_code == 404:
                raise NotFound(response)
            if response.status_code == 429:
                if self.wait_on_rate_limit:
                    reset_time = int(response.headers["x-rate-limit-reset"])
                    sleep_time = reset_time - int(time.time()) + 1
                    if sleep_time > 0:
                        log.warning(
                            "Rate limit exceeded. "
                            f"Sleeping for {sleep_time} seconds."
                        )
                        time.sleep(sleep_time)
                    return self.request(method, route, params, json_data, auth, content_type, accept)
                else:
                    raise TooManyRequests(response)
            if response.status_code >= 500:
                raise ServerError(response)
            if not 200 <= response.status_code < 300:
                raise HTTPException(response)

            return response

    def _make_request(self, method, route, params={},
                      json_data=None, auth=None, content_type=None, accept=None):

        response = self.request(method, route, params=params,
                                json_data=json_data, auth=auth, content_type=content_type, accept=accept)

        # if return_type is dict and accept == "application/json":
        return response.json()
        # else:
        #    return response


class DdcClient(BaseClient):
    """
    Represents a Danube Data Cube client.

    Args:
        client_id (str): Danube Data Cube client id.
        client_secret (str): Danube Data Cube client secret.
        host (Optional[str] ): Alternative Danube Data Cube host url.

    Examples:
        # Create a DDC Client object for interacting with the DDC service
        >>> client = DdcClient(<client_id>, <client_secret>)

        # List the user's AOIs
        >>> DDC.get_all_aoi()

    """

    def __init__(self,
                 client_id: Optional[str] = None,
                 client_secret: Optional[str] = None,
                 host: Optional[str] = None) -> None:

        client_id = client_id or os.environ.get('DDC_CLIENT_ID')
        client_secret = client_secret or os.environ.get('DDC_CLIENT_SECRET')
        host = host or DEFAULT_DDC_HOST

        if not client_id or not client_secret:
            raise DdcException(
                'Both "client_id" and "client_secret" must be provided. '
                'Consider setting environment variables '
                'DDC_CLIENT_ID and DDC_CLIENT_SECRET.'
            )

        self.auth = None
        self.aws_s3 = None
        self.aws_session_exp = 0
        
        super().__init__(client_id, client_secret, host, False)

    @authorize_request
    def get_all_aoi(self, with_geometry: bool = True, output_data_type: int = 1) -> Union[pd.DataFrame, List[Dict]]:
        """
        Get user's all area of interests (AOI).

        Args:
            with_geometry (bool, optional): Whether to retrive geometry values.
                Defaults to True.
            output_data_type (int, optional): Whether to return the result as pandas.DataFrame (1) or as list of dictionaries (2).
                Defaults to 1.

        Returns:
            Union[pd.DataFrame, List[Dict]]: Depending on the `output_data_type` parameter,
            this function returns either a pandas DataFrame or a list of dictionary containing AOI information.

        Raises:
            DdcRequestError: If an error occurs during the process of requesting AOIs.
        """

        params = {'user_id': self.client_id,
                  'with_geometry': with_geometry}

        route = "/aoi_manager/get_aoi"
        accept = "application/json"

        try:
            response = self._make_request(
                "GET", route, params=params, auth=self.auth, accept=accept)

        except HTTPException as error:
            raise DdcRequestError(
                f"Error during getting all aoi with params {params} from {route}"
            ) from error

        return self._process_response(response, output_data_type)

    @authorize_request
    def get_aoi_by_id(self, aoi_id: int, output_data_type: int = 1) -> Union[pd.DataFrame, Dict]:
        """
        Get user's area of interests (AOI) by ID.

        Args:
            aoi_id (int): ID of the AOI.
            output_data_type (int, optional): Whether to return the result as pandas.DataFrame (1) or as a dictionary(2).
                Defaults to 1.

        Returns:
            Union[pd.DataFrame, Dict]: Depending on the `output_data_type` parameter,
            this function returns either a pandas DataFrame or a dictionary containing AOI information.

        Raises:
            DdcRequestError: If an error occurs during the process of requesting the AOI.
        """

        params = {'user_id': self.client_id,
                  'aoi_id': aoi_id}

        route = "/aoi_manager/get_aoi"
        accept = "application/json"

        try:
            response = self._make_request(
                "GET", route, params=params, auth=self.auth, accept=accept)

        except HTTPException as error:
            raise DdcRequestError(
                f"Error during getting all aoi with params {params} from {route}"
            ) from error

        return self._process_response(response, output_data_type)

    @authorize_request
    def create_aoi(self,
                   name: str,
                   geometry: Union[Geometry, Polygon, MultiPolygon, str],
                   time_range: Union[TimeRange, Tuple[pd.Timestamp, pd.Timestamp], Tuple[str, str]],
                   data_layers: Union[List[DataCollection], List[Dict]] = None,
                   layer_selection_id:  int = None,
                   layer_ids: List[int] = [],
                   is_dynamic: bool = False,
                   output_data_type: int = 1
                   ) -> Union[pd.DataFrame, Dict]:
        """
        Create an area of interests (AOI).

        Args:
            name (str): The name of the area of interest.
            geometry (Union[Geometry, Polygon, MultiPolygon, str]): The geometry of the area of interest.
                This can be provided as a `ddc_utility.Geometry` object, a `shapely.Polygon`, a `shapely.MultiPolygon`, or as a WKT or GeoJSON string.
            time_range (Union[TimeRange, Tuple[pd.Timestamp, pd.Timestamp], Tuple[str, str]]):
                The time range for which the area of interest is defined.
                This can be provided as a `ddc_utility.TimeRange` object, a tuple of two `pandas.Timestamp` objects,
                or a tuple of two strings representing dates.
            data_layers (Union[List[DataCollection], List[Dict]], optional): A list of dictionaries representing data layers to associate with the AOI.
                Each dictionary should contain information about a specific data layer.
                Example:
                    [
                        {'name': 'S2L2A', 'bands': ['B04', 'B08']},
                        {'name': 'Meteorology', 'bands': ['Temp Avg', 'Prec']}
                    ]
                If None, the AOI cube will be created with an empty polygon.
                Defaults to None.
            layer_selection_id (int): Layer selection ID. 
            layer_ids (List[int]): List of layer IDs.
            is_dynamic (bool, optional): Whether the AOI is dynamic (True) or static (False).
                Defaults to False.
            output_data_type (int, optional): Whether to return the result as a `pandas.DataFrame` (1) or as a dictionary (2).
                Defaults to 1.

        Returns:
            Union[pd.DataFrame, Dict]: Depending on the `output_data_type` parameter,
            this function returns either a pandas DataFrame or a dictionary containing AOI information.

        Raises:
            DdcRequestError: If an error occurs during the process of creating the AOI.

        """
        if not isinstance(time_range, TimeRange):
            time_range = TimeRange(*time_range)
        time_range_str = time_range.to_string(only_date=True)

        if not isinstance(geometry, Geometry):
            geometry = Geometry(geometry)
        geometry_str = geometry.to_string()

        if data_layers is None:
            data_layers_lst = []
        else:
            layers = []
            for layer in data_layers:
                if not isinstance(layer, DataCollection):
                    layer = DataCollection(
                        name=layer['name'], bands=layer['bands'])
                layers.append(layer)
            data_layers_lst = prepare_data_layers(layers)

        json_data = {
            "user_id": self.client_id,
            "name": name,
            "geometry": geometry_str,
            "start_date": time_range_str[0],
            "end_date": time_range_str[1],
            "is_dynamic": is_dynamic
        }

        if layer_selection_id:
            json_data["layer_selection_id"] = layer_selection_id
        elif len(layer_ids) > 0:
            layer_ids_str = ','.join(str(x) for x in layer_ids)
            json_data["layer_ids"] = layer_ids_str
        else:
            json_data["data_layers"] = {"data_sources": data_layers_lst}

        route = "/aoi_manager/create_aoi"
        accept = "application/json"
        content_type = "application/json"

        try:
            response = self._make_request(
                "POST", route, json_data=json_data, auth=self.auth, content_type=content_type, accept=accept)

        except HTTPException as error:
            raise DdcRequestError(
                f"Error during creating aoi with body {json_data} from {route}"
            ) from error

        return self._process_response(response, output_data_type)

    @authorize_request
    def get_data_layers(self, output_data_type: int = 1) -> Union[pd.DataFrame, List[Dict]]:
        """
        Get available data layers

        Args:

        Returns:
            Union[pd.DataFrame, Dict]: Depending on the `output_data_type` parameter,
            this function returns either a pandas DataFrame or a dictionary containing crop type information.
        """

        params = {'user_id': self.client_id}

        route = "/aoi_manager/data_layers"
        accept = "application/json"

        try:
            response = self._make_request(
                "GET", route, params=params, auth=self.auth, accept=accept)

        except HTTPException as error:
            raise DdcRequestError(
                f"Error during getting crop types with params {params} from {route}"
            ) from error

        return self._process_response(response, output_data_type)

    @authorize_request
    def get_crop_types(self, output_data_type: int = 1) -> Union[pd.DataFrame, List[Dict]]:
        """
        Get available crop types.

        Args:
            output_data_type (int, optional): Whether to return the result as pandas.DataFrame (1) or as a dictionary(2).
                Defaults to 1.

        Returns:
            Union[pd.DataFrame, List[Dict]]: Depending on the `output_data_type` parameter,
            this function returns either a pandas DataFrame or a list of dictionary containing crop type information.

        Raises:
            DdcRequestError: If an error occurs during the process of requesting crop types.
        """

        params = {'user_id': self.client_id}

        route = "/crop/get_type"
        accept = "application/json"

        try:
            response = self._make_request(
                "GET", route, params=params, auth=self.auth, accept=accept)

        except HTTPException as error:
            raise DdcRequestError(
                f"Error during getting crop types with params {params} from {route}"
            ) from error

        return self._process_response(response, output_data_type)

    @authorize_request
    def create_crop_type(self,
                         crop_type_name: str,
                         output_data_type: int = 1
                         ) -> Union[pd.DataFrame, Dict]:
        """
        Create crop type.

        Args:
            crop_type_name (str): Name of the crop type.
            output_data_type (int, optional): Whether to return the result as pandas.DataFrame (1) or as a dictionary(2).
                Defaults to 1.

        Returns:
            Union[pd.DataFrame, Dict]: Depending on the `output_data_type` parameter,
            this function returns either a pandas DataFrame or a dictionary containing crop type information.

        Raises:
            DdcRequestError: If an error occurs during the process of creating the growing season.

        """

        json_data = {
            "user_id": self.client_id,
            "crop_type_name": crop_type_name
        }

        route = "/crop/create_type"
        accept = "application/json"
        content_type = "application/json"

        try:
            response = self._make_request(
                "POST", route, json_data=json_data, auth=self.auth, content_type=content_type, accept=accept)

        except HTTPException as error:
            raise DdcRequestError(
                f"Error during creating crop type with body {json_data} from {route}"
            ) from error

        return self._process_response(response, output_data_type)

    @authorize_request
    def get_crop_variety(self, crop_type_id: int, output_data_type: int = 1) -> Union[pd.DataFrame, List[Dict]]:
        """
        Get available crop types.

        Args:
            crop_type_id (int): ID of crop type.
            output_data_type (int, optional): Whether to return the result as pandas.DataFrame (1) or as a dictionary(2).
                Defaults to 1.

        Returns:
            Union[pd.DataFrame, List[Dict]]: Depending on the `output_data_type` parameter,
            this function returns either a pandas DataFrame or a list of dictionary containing crop variety information.

        Raises:
            DdcRequestError: If an error occurs during the process of requesting crop variety.
        """

        params = {'user_id': self.client_id,
                  'crop_type_id': crop_type_id}

        route = "/crop/get_variety"
        accept = "application/json"

        try:
            response = self._make_request(
                "GET", route, params=params, auth=self.auth, accept=accept)

        except HTTPException as error:
            raise DdcRequestError(
                f"Error during getting crop variety with params {params} from {route}"
            ) from error

        return self._process_response(response, output_data_type)

    @authorize_request
    def create_crop_variety(self,
                            crop_type_id: id,
                            crop_variety_name: str,
                            output_data_type: int = 1
                            ) -> Union[pd.DataFrame, Dict]:
        """
        Create crop variety.

        Args:
            crop_type_id (id): ID of crop type.
            crop_variety_name (str): Name of the crop variety.
            output_data_type (int, optional): Whether to return the result as pandas.DataFrame (1) or as a dictionary(2).
                Defaults to 1.

        Returns:
            Union[pd.DataFrame, Dict]: Depending on the `output_data_type` parameter,
            this function returns either a pandas DataFrame or a dictionary containing crop variety information.

        Raises:
            DdcRequestError: If an error occurs during the process of creating the crop variety.

        """

        json_data = {
            "user_id": self.client_id,
            "crop_type_id": crop_type_id,
            "crop_variety_name": crop_variety_name
        }

        route = "/crop/create_variety"
        accept = "application/json"
        content_type = "application/json"

        try:
            response = self._make_request(
                "POST", route, json_data=json_data, auth=self.auth, content_type=content_type, accept=accept)

        except HTTPException as error:
            raise DdcRequestError(
                f"Error during creating crop variety with body {json_data} from {route}"
            ) from error

        return self._process_response(response, output_data_type)

    @authorize_request
    def get_crop_models(self, crop_type_id: int, output_data_type: int = 1) -> Union[pd.DataFrame, List[Dict]]:
        """
        Get available crop model list.

        Args:
            crop_type_id (int): ID of crop type.
            output_data_type (int, optional): Whether to return the result as pandas.DataFrame (1) or as a dictionary(2).
                Defaults to 1.

        Returns:
            Union[pd.DataFrame, List[Dict]]: Depending on the `output_data_type` parameter,
            this function returns either a pandas DataFrame or a list of dictionary containing crop model information.

        Raises:
            DdcRequestError: If an error occurs during the process of requesting crop models.
        """

        params = {'user_id': self.client_id,
                  'crop_type_id': crop_type_id}

        route = "/crop/get_model"
        accept = "application/json"

        try:
            response = self._make_request(
                "GET", route, params=params, auth=self.auth, accept=accept)

        except HTTPException as error:
            raise DdcRequestError(
                f"Error during getting crop models with params {params} from {route}"
            ) from error

        return self._process_response(response, output_data_type)

    @authorize_request
    def run_crop_model(self,
                       aoi_id: int,
                       time_range: Union[TimeRange, Tuple[pd.Timestamp, pd.Timestamp], Tuple[str, str]],
                       sowing_date: Union[pd.Timestamp, str],
                       irrigation: List[Union[IrrigationSchedule, Tuple[Union[pd.Timestamp, str], float]]],
                       init_water_content: float,
                       growing_season_id: int,
                       crop_model_name: str,
                       seasonal_trajectory: bool,
                       soil_type: str,
                       output_data_type: int = 1) -> Union[pd.DataFrame, List[Dict]]:
        """
        Run crop model.

        Args:
            aoi_id (int): ID of the AOI.
            time_range (Union[TimeRange, Tuple[pd.Timestamp, pd.Timestamp], Tuple[str, str]]): Time range for the simulation.
            sowing_date (Union[pd.Timestamp, str]): Sowing date for the simulation.
            irrigation (List[Union[IrrigationSchedule, Tuple[Union[pd.Timestamp, str], float]]]): Irrigation schedule for the simulation.
            init_water_content (float): Initial water content for the simulation.
            growing_season_id (int): ID of the growing season.
            crop_model_name (str): Name of the crop model.
            seasonal_trajectory (bool): Flag for utilization of CLMS PPI ST in the modelling process
            soil_type (str): USDA soil type definition  
            output_data_type (int, optional): Whether to return the result as pandas.DataFrame (1) or as a dictionary(2).
                Defaults to 1.

        Returns:
            Union[pd.DataFrame, List[Dict]]: Depending on the `output_data_type` parameter,
            this function returns either a pandas DataFrame or a list of dictionary containing crop model information.

        Raises:
            DdcRequestError: If an error occurs during the process of running crop model.

        """

        if not isinstance(time_range, TimeRange):
            time_range = TimeRange(*time_range)
        time_range_str = time_range.to_string(only_date=True)

        if not isinstance(sowing_date, pd.Timestamp):
            sowing_date = pd.Timestamp(sowing_date)
        sowing_date_str = sowing_date.isoformat(sep='T').split('T')[0]

        irrigation_str = []
        for irr in irrigation:
            if not isinstance(irr, IrrigationSchedule):
                irr = IrrigationSchedule(*irr)
            irrigation_str.append(irr.to_string(only_date=True))

        params = {
            "user_id": self.client_id,
            "aoi_id": aoi_id,
            "start_date": time_range_str[0],
            "end_date": time_range_str[1],
            "sowing_date": sowing_date_str,
            "irrigation": irrigation_str,
            "init_water_content": init_water_content,
            "growing_season_id": growing_season_id,
            "crop_model_name": crop_model_name,
            "seasonal_trajectory": seasonal_trajectory,
            "soil_type": soil_type
        }

        route = "/crop_model/run"
        accept = "application/json"

        try:
            response = self._make_request(
                "GET", route, params=params, auth=self.auth, accept=accept)

        except HTTPException as error:
            raise DdcRequestError(
                f"Error during running crop modle with params {params} from {route}"
            ) from error

        return self._process_response(response, output_data_type)

    @authorize_request
    def get_growing_season(self, aoi_id: int, output_data_type: int = 1) -> Union[pd.DataFrame, List[Dict]]:
        """
        Get growing seasons for AOI.

        Args:
            aoi_id (int): ID of the AOI.
            output_data_type (int, optional): Whether to return the result as pandas.DataFrame (1) or as a dictionary(2).
                Defaults to 1.

        Returns:
            Union[pd.DataFrame, List[Dict]]: Depending on the `output_data_type` parameter,
            this function returns either a pandas DataFrame or a list of dictionary containing growing season information.

        Raises:
            DdcRequestError: If an error occurs during the process of requesting growing seasons.
        """

        params = {'user_id': self.client_id,
                  'aoi_id': aoi_id}

        route = "/growing_season/get_season"
        accept = "application/json"

        try:
            response = self._make_request(
                "GET", route, params=params, auth=self.auth, accept=accept)

        except HTTPException as error:
            raise DdcRequestError(
                f"Error during getting growing seasons with params {params} from {route}"
            ) from error

        return self._process_response(response, output_data_type)

    @authorize_request
    def create_growing_season(self,
                              aoi_id: int,
                              time_range: Union[TimeRange, Tuple[pd.Timestamp, pd.Timestamp], Tuple[str, str]],
                              sowing_date: Union[pd.Timestamp, str],
                              crop_type_id: int,
                              crop_variety_id: int,
                              crop_model_id: int,
                              output_data_type: int = 1
                              ) -> Union[pd.DataFrame, Dict]:
        """
        Create growing season for AOI.

        Args:
            aoi_id (int): ID of the AOI.
            time_range (Union[TimeRange, Tuple[pd.Timestamp, pd.Timestamp], Tuple[str, str]]):
                The time range for which the growing season is defined.
                This can be provided as a `ddc_utility.TimeRange` object, a tuple of two `pandas.Timestamp` objects,
                or a tuple of two strings representing dates.
            sowing_date (Union[pd.Timestamp, str]): The date when the crop is sown.
            crop_type_id (int): ID of crop type.
            crop_variety_id (int): ID of crop variety.
            crop_model_id (int): ID of crop model.
            output_data_type (int, optional): Whether to return the result as pandas.DataFrame (1) or as a dictionary(2).
                Defaults to 1.

        Returns:
            Union[pd.DataFrame, Dict]: Depending on the `output_data_type` parameter,
            this function returns either a pandas DataFrame or a dictionary containing growing season information.

        Raises:
            DdcRequestError: If an error occurs during the process of creating the growing season.

        """

        if not isinstance(time_range, TimeRange):
            time_range = TimeRange(*time_range)
        time_range_str = time_range.to_string(only_date=True)

        if not isinstance(sowing_date, pd.Timestamp):
            sowing_date = pd.Timestamp(sowing_date)
        sowing_date_str = sowing_date.isoformat(sep='T').split('T')[0]

        json_data = {
            "user_id": self.client_id,
            "aoi_id": aoi_id,
            "start_date": time_range_str[0],
            "end_date": time_range_str[1],
            "sowing_date": sowing_date_str,
            "crop_type_id": crop_type_id,
            "crop_variety_id": crop_variety_id,
            "crop_model_id": crop_model_id
        }

        route = "/growing_season/create_season"
        accept = "application/json"
        content_type = "application/json"

        try:
            response = self._make_request(
                "POST", route, json_data=json_data, auth=self.auth, content_type=content_type, accept=accept)

        except HTTPException as error:
            raise DdcRequestError(
                f"Error during creating growing season with body {json_data} from {route}"
            ) from error

        return self._process_response(response, output_data_type)

    @authorize_s3_access
    def open_aoi_cube(self, aoi_id: int, bucket_name: str = DEFAULT_AOI_BUCKET) -> xr.Dataset:
        """
        Open AOI cube as an xarray.Dataset.

        Args:
            aoi_id (int): ID of the AOI.
            bucket_name (str, optional): Name of the S3 bucket where the zarr cube is stored.
                Defaults to `DEFAULT_AOI_BUCKET`.

        Returns:
            xr.Dataset: AOI cube.

        Raises:
            DdcClientError: If user don't have access to the bucket.
            DdcRequestError: If an error occurs during opening the cube.

        """

        zarr_path = f"s3://{bucket_name}/{aoi_id}_{self.client_id}.zarr"

        try:
            cube = open_cube(path=zarr_path, fs=self.aws_s3)
        except PermissionError as error:
            raise DdcClientError(
                "User don't have access for this operation") from error
        except FileNotFoundError as error:
            raise DdcRequestError(
                f"Invalid aoi_id - no such aoi cube under{zarr_path}") from error
        except Exception as error:
            raise DdcRequestError(
                f"Error during getting AOI cube from {zarr_path}"
            ) from error
        return cube

    @authorize_s3_access
    def open_ddc_cube(self, zarr_path: str, zarr_group: Optional[str] = None, bucket_name: str = DEFAULT_DDC_BUCKET) -> xr.Dataset:
        """
        Open DDC dataset as an xarray.Dataset.

        Args:
            zarr_path (str): zarr path to the dataset.
            zarr_group (str): zarr group of the dataset.
            bucket_name (str, optional): Name of the S3 bucket where the zarr cube is stored.
                Defaults to `DEFAULT_AOI_BUCKET`.

        Returns:
            xr.Dataset: DDC cube.

        Raises:
            DdcClientError: If user don't have access to the bucket.
            DdcRequestError: If an error occurs during opening the cube.

        """

        zarr_path = f"s3://{bucket_name}/{zarr_path}"

        try:
            cube = open_cube(path=zarr_path,
                             fs=self.aws_s3,
                             group=zarr_group)
        except PermissionError as error:
            raise DdcClientError(
                "User don't have access for this operation") from error
        except Exception as error:
            raise DdcRequestError(
                f"Error during getting DDC cube from {zarr_path}"
            ) from error
        return cube

    
    def _process_response(self, data: Any, output_data_type: int = None):
        if output_data_type == 1:
            if isinstance(data, list):
                return pd.DataFrame(data)
            if isinstance(data, dict):
                return pd.DataFrame([data])
            else:
                raise ValueError(
                    f"Can't post-process API response -- {type(data)} is invalid with output_data_type of {output_data_type}")
        else:
            return data

    def fetch_token(self) -> Dict:
        """Fetch token from a remote token endpoint."""

        route = "/get_token"
        accept = "application/json"
        content_type = "application/json"
        json_data = {
            "client_id": self.client_id,
            "client_secret": self.client_secret
        }

        now = time.time()

        try:
            response = self._make_request(
                "POST", route, json_data=json_data, content_type=content_type, accept=accept)

        except HTTPException as error:
            raise DdcRequestError(
                f"Error during fetching token from {route}"
            ) from error

        if "Error" in response:
            raise DdcClientError(
                f"Error during fetching token for user: {response['Error']}")

        if response['token_type'].lower() != 'bearer':
            raise DdcRequestError(
                f"Expected token_type to equal 'bearer',but got {response['token_type']} instead")

        response['expires_at'] = now + int(response.pop('expires_in'))

        return response

    @authorize_request
    def fetch_temporary_credentials(self) -> Dict:
        """Fetch token from a remote token endpoint."""

        route = "/get_temp_cred"
        accept = "application/json"
        params = {'user_id': self.client_id}

        try:
            response = self._make_request(
                "GET", route, params=params, auth=self.auth, accept=accept)

        except HTTPException as error:
            raise DdcRequestError(
                f"Error during fetching temporary credential from {route}"
            ) from error

        response = response["Credentials"]

        response["expires_at"] = int(time.mktime(time.strptime(
            response.pop("Expiration"), "%Y-%m-%dT%H:%M:%SZ")))

        return response
